
# coding: utf-8

# # Whatever Forever
# 
# Pythonic syntaxes that save pixels when developing in the notebook.

# ## `Chain`
# 
# A chain is a typographically compact manner of creating complicated
# expressions in Pythonic syntax.
# 
# ### Chainable Values
# 
# ```python
# some_expr = Chain([1,2,3]).reversed().map(lambda x: x**2).list()
# some_expr.value()
# some_expr.value([3,5,8])
# ```
# 
# ### Syntactic Sugar
# 
# `\` & `>` offer chain new functions and evaluate them respectively.
# 
# ```python
# from toolz.curried import *
# some_expr = Chain([1,2,3]) | reversed | map(lambda x: x**2) | list
# some_value = Chain([1,2,3]) | reversed | map(lambda x: x**2) > list
# ```

# ## `Whatever`
# 
# Easy to construct cell magics
# 
# ### Cell Magics
# 
# Create a `jinja` to Markdown magic.
# 
# ```python
# from whatever4e import Forever
# from jinja2 import Template
# @Forever.cell('jinja2', lang='jinja2', display='Markdown')
# def render_jinja_with_globals(cell):
#     return Template(cell).render(**globals())
# ```

# ## License
# `whatever4e` is released as free software under the [BSD 3-Clause license]
# (https://github.com/tonyfast/whatever-forever/blob/master/LICENSE).

# In[ ]:

__version_info__ = (0, 0, 8)
__version__ = '.'.join(map(str, __version_info__))

from .chain import Chain
from .magic import Forever
from class_maker import method

__all__ = [
    'Forever', 'Chain', 'method',
]




from toolz.curried import curry
from inspect import isclass
from types import MethodType

__all__ = [
    'add_method',
]

@curry
def method(
    class_or_instance,
    f,
    decorators = [],
):
    """Add a method to class or instance.
    """
    name, fn = f.__name__, f
    if not isclass(class_or_instance):
        fn = MethodType(fn, class_or_instance)
    for decorator in decorators:
        fn = decorator(fn)
    setattr(class_or_instance, name, fn)
    return f



# coding: utf-8

# In[2]:

from IPython import display, get_ipython
from IPython.core import magic_arguments
from IPython.core.magic import (
    Magics,
    magics_class,
    line_cell_magic,
)
from typing import Callable
from toolz.curried import *


# > Evaluate arbitrary variables that can be added to the global context by defining a name.

# In[21]:

@magic_arguments.magic_arguments()
@magic_arguments.argument(
    "name",
    default=None,
    nargs="?",
    help="""Name of local variable to set to parsed value""",
)
@magic_arguments.argument(
    "-d",
    "--display",
    default='Markdown',
    nargs="?",
    help="""An IPython.display method."""
)
def magical_cell(line, cell, f=identity, **kwargs):
    args = magic_arguments.parse_argstring(magical_cell, line.strip())

    retval = f(cell)
    
    if args.name:
        if '.' in args.name:
            path = args.name.split('.')
            var = get_ipython().user_ns[path[0]]
            setattr( reduce(
                lambda x, y: getattr(x,y),
                path[1:-1], 
                var
            ), path[-1], retval)
        else:   
            get_ipython().user_ns[args.name] = retval
    
    
    if args.display:
        disp = kwargs['display'] if 'display' in kwargs else args.display
        if isinstance(disp, str):
            return display.display(getattr(display, disp)(retval))
        elif isinstance(disp, Callable):
            return disp(retval)


# In[4]:

class magical():
    def __init__(self, magic_kind=magical_cell):
        self._magic_kind = magic_kind
    
    @curry
    def __call__(self, key, f, lang=None, **kwargs):
        name = self._magic_kind.__name__.split('_')[1]
        get_ipython().register_magic_function(
            partial(self._magic_kind, f=f, **kwargs),
            magic_kind = name,
            magic_name = key,
        )
        if lang:
            # Syntax highlighting
            pipe("""require([
                        "notebook/js/codecell",
                        "codemirror/mode/{0}/{0}"
                    ],
                    function(cc){{
                        cc.CodeCell.options_default.highlight_modes.magic_{1} = {{
                            reg: ["^%%{1}"]
                        }};
                    }}
                );
                """.format(lang, name), 
                 display.Javascript, display.display,
            )


# > Still don't know how I will use this.

# In[5]:

def magical_line(line, f=identity, **kwargs):    
    """I don't understand how I would use this yet."""
    if 'assign' in kwargs:
        if kwargs['assign']:
            line, cell = line.strip().split(' ',1)
    else:
        line, cell = ['', line]
    return magical_cell(line, cell, f, **kwargs)


# In[6]:

class Forever(object):
    cell = staticmethod(magical(magical_cell))
    line = staticmethod(magical(magical_line))





# coding: utf-8

# It should be easier to construct complicated functions.  Python has named a lot of things, and
# large scale adoption of certain libraries make names meaningful.  `Chain` is inspired by `pytoolz`
# and `underscorejs`.  The [`pytoolz` documentation does a fantastic describing functional programming](toolz.readthedocs.io).
# 
# There are a few functional programming tools in Python ecosystem.  I would up breaking them 
# very quickly after trying them __Or__ the didn't work in the Jupyter notebook.
# 
# A `Chain` is efficient for prototyping and scaling complicated actions.  
# 
# A Chain class
# 
# * composes functions.  `Chain` must be explicitly computed using `value` or `>`
# * has a special attribute `value`
# * a `copy` method
# * overloads `__or__` and `__gt__`
#     * Provides semi-functional syntax for evaluation

# The only required import is `toolz` from pypi

# In[1]:

from class_maker import method
import builtins
import toolz.curried
from toolz.curried import *
from types import MethodType
from typing import Callable


# In[10]:

class Chain(object): 
    _imports = [toolz.curried, builtins]
    
new_method = method(Chain)


# In[17]:

@new_method
def __init__(
    self,
    *args, **kwargs
):
    # A function to extract attributes from a module.
    module_methods = lambda module: pipe(module, dir, filter(
            lambda x: not x.startswith('_')
        ), map(
            partial(getattr, module)
        ), list)
    
    # Create a list of attributes that can be added to chain
    # It would be cool to have some code prediction.
    self._imports = pipe(self._imports, map(module_methods), concat, filter(lambda x: hasattr(x,'__name__')),list)
    
    # Tokens record function, arguments, and keywork arguments.
    self._tokens = []
    
    # Default context to evaluate the chain.
    self._args = args
    self._kwargs = kwargs


# > Method to compose a function from the tokens that 
# have been created in the chain

# In[18]:

@new_method
def _compose(self):
    """From the tokens, compose a function that can be reused.
    """
    return compose(
        *pipe(
            self._tokens if self._tokens else [[identity, [], {}]],
            reversed,
            map(
                lambda e: partial(
                    e[0], *e[1], **e[2]
                ) if e[1] or e[2] else e[0],
            )
        )
    )


# > `value` composes the function and then evaluate it with either 
# the default context or a new set of arguments to evaluate with the same composition.

# In[19]:

@new_method
def value(self, *args, **kwargs):
    """Compose and evaluate the function.  If no args or kwargs are provided
    then the initialized context is used.
    """
    fn = self._compose()
    if args or kwargs:
        return fn(*args, **kwargs)
    if self._kwargs: 
        return fn(*self._args, **self._kwargs)
    return fn(*self._args)


# > `__call__` does not execute anything it just records args 
# and kwargs that would be passed to the function.

# In[24]:

@new_method
def __call__(self, *args, **kwargs):
    self._tokens[-1][1] = args
    self._tokens[-1][2] = kwargs
    return self


# In[25]:

@new_method
def _getter(self, attr):
    """Extract a function from the imports based on a name.
    """
    return pipe(self._imports, filter(lambda x: x.__name__ == attr), first)


# ```python
# Chain([1,2,3,4]).reversed().str()
# ```

# In[30]:

@new_method
def __getattr__(self, attr):
    """Add a token for an attribute in _imports.
    """
    self._tokens.append([self._getter(attr), [], {}])
    return self


# > A mighty flexible way to append any function the chain.

# ```python
# Chain([1,2,3,4])[reversed][str]
# ```

# In[127]:

@new_method
def __getitem__(self, item):
    self._tokens.append([item, (), {}])
    return self


# > A `Chain` is a `Chain` until `value` or `|`.  

# > Evaluate when the repr is called.

# In[129]:

@new_method
def __repr__(self):
    return self.value().__str__()


# > Copy the state of the chain and return a new one.

# In[130]:

@new_method
def copy(self):
    chain = Chain(*self._args, **self._kwargs)
    chain._imports = self._imports[:]
    chain._tokens = self._tokens[:]
    return chain   


# > Some sweet ol' syntactic sugar that looks like functional programming.
# 
# > `__or__` adds a piping method to a chain.  Any functional can be included in the pipe.
# 
# > `__or__` is shorthand for a copy and a an append.

# In[31]:

@new_method
def __or__(self, f):
    chain = self.copy()
    chain._tokens.append([f, [], {}])
    return chain


# > `__gt__` is shorthand for a copy, append, and a `value` operation.

# In[132]:

@new_method
def __gt__(self, f):
    return self.copy()[f].value()


# # About the design
# 
# The goal of `Chain` is rapidly prototype code using powerful modules in the 
# Python ecosystem.  More specifically, this was designed with the intent of 
# creating data rich web applications that may work client or server side.
# 
# Applications are complex.  Fortunately, `conda` and `pip` provide lower 
# barriers to entry for building complex systems.  With this access, it is 
# possible to write functional code that spans loading through an ETL pipeline
# to an application.  This workflow was inspired greatly by `toolz` and chainable `d3`
# expressions.  Choosing these opinions greatly minimizes the need for naming.
# 
# ## Naming
# 
# Naming takes time and have the potential to confuse others.  The python ecosystem 
# provides a broad corpus to load, transform, model, and visualize data with simple
# methods to deploy them into applications or scale to other day.
# 
# I don't want to name stuff.  Also, I want to write code quicker.  Using existing namespaces
# enhances code prediction while developing in the notebook.
# 
# ## Intent
# 
# * Design a lightweight class that offers a typographically and functionally compact python syntax.
# * Maximize screen real estate for impact.  This optimizes copy and paste ability.
# * Lazy.  Most of this comes from toolz.
# * Reusability.
# * Don't create a new grammar.
# 
# ## Advantages
# 
# * Less decisions to make, better code prediction.
# * Complex operations can be isolated to a notebook cell
# * Enhanced copy and pasting abilities.
# 
# ## History
# 
# 1. Design an API that mimics [`chain` from `underscorejs`](http://underscorejs.org/#chain).  This 
# includes the `value` function that evaluates that `chain`.  `Chain` could only use `builtins` and `toolz.curried`.
# 
#     ```python
#     Chain([1,2,3]).map(lambda x: x**2).list()
#     ```
#     
#     `value` can take args and kwargs to apply a different context to the functions.
# 
#     ```python
#     c = Chain([1,2,3]).map(lambda x: x**2).list()
#     c.value([3,5,7])
#     ```
# 
# 2. Extend the API to functions in the global namespace using the `__getitem__` method.  Chaining functions
# using the attribute created a nice typographic layout.
# 
#     ```python
#     Chain([1,2,3])[
#         map(lambda x: x**2)
#     ].list()
#     ```
#     
# 3. Add syntactic sugar to extend and evaluate a chain.  
# 
#     __extend a chain__ `Chain([1,2,3]) | map(lambda x: x**2) | list`
#     
#     __evaluate a chain__ `Chain([1,2,3]) | map(lambda x: x**2) > list`

# In[ ]:



